/* 
 * This is to write out to the disk in the case that we have
 * essentially a non-functional TiVo.  The idea here is we're
 * going to write out ot the swap partition some log of what
 * happened.  The two cases to get us here generally are
 * a kernel panic, or a watch dog fire.
 *
 * We are going to write out to swap, on boot up, we'll scan
 * the swap before turning it on to see our signature.  If
 * the signature is present, then we'll extract the data
 * and put it in our normal logs.
 *
 * Copyright TiVo (C) 2005
 * Written by Shane Nay (snay@tivo.com)
 *
 */

#include <linux/init.h>
#include <linux/module.h>
#include <linux/types.h>
#include <linux/notifier.h>
#include <linux/ide-deadsystem.h>
#include <linux/string.h>
#include <linux/paniclog.h>
#include <linux/plog.h>

#define PLOG_CAPTURE_SIZE      (8*1024)     /* 8k seems like a fun number */

int pio_write_plog(int secoffset, u32 *csum) {
	/*
	 * So, this is kind of strange.  We start at the end,
	 * and iterate backwards.  So, we'll write our last
	 * sector first, and our first sector last.
	 */

#if defined(CONFIG_PLOG)

	int cursect = secoffset + (PLOG_CAPTURE_SIZE-512)/512;
	u8 secbuf[512];
	int ret;
	
	while(cursect >= secoffset) {
		plog_destructive_liforead((int*)secbuf, 512/4);
		ret=pio_write_buffer_to_swap(secbuf, 512, cursect, csum);
		if(ret < 0)
			return ret;
		cursect--;
	}
	return PLOG_CAPTURE_SIZE/512;
#else
	return 0;
#endif
}

#if defined(CONFIG_MODULES)

#define MAX_SYMNAME_SIZE 120

int calc_sym_size(struct module_symbol *sym) {
	int size;
	for(size=0;size < MAX_SYMNAME_SIZE && sym->name[size] != '\0';size++);
	return size + 4 + 1;
}
#endif

int pio_write_syms(int secoffset, u32 *csum) {
#if defined(CONFIG_MODULES)

	/*
	 * Iterate through modules, printing out binary data
	 * when we flow over the buffer.  Oh to have an unlimited
	 * stack.  Hrmph.
	 */


	int wrsects = 0;
	u8 secbuf[512];
	int symsizeused=0;

	struct module *mod;

	for(mod = module_list; mod; mod = mod->next) {
		int i;
		struct module_symbol *sym;
		
		if(!MOD_CAN_QUERY(mod))
			continue;
		
		for(i = mod->nsyms, sym = mod->syms; i > 0; --i , ++sym) {
			int symsize = calc_sym_size(sym);

			if(symsizeused + symsize > 512) {
				/* Time to flush out the secbuf */
				if(symsizeused < 512) {
					memset(secbuf + symsizeused, 0,
					       512 - symsizeused);
				}
				pio_write_buffer_to_swap(secbuf, 512,
					       wrsects++ + secoffset, csum);
				symsizeused=0;
			}
			memcpy((void*)(secbuf + symsizeused), &sym->value, 4);
			memcpy((void*)(secbuf + symsizeused + 4), sym->name,
			       symsize - 5);

			secbuf[symsize + symsizeused - 1] = '\0';
			symsizeused += symsize;
		}
	}

	if(symsizeused > 0) {
		if(symsizeused < 512) {
			memset(secbuf + symsizeused, 0, 512 - symsizeused);
		}
		pio_write_buffer_to_swap(secbuf, 512, wrsects++ + secoffset,
					 csum);
	}

	return wrsects;

#else
	return 0;
#endif
}

extern char *linux_banner;

int pio_write_header(int syslogsize, int plogsize,
		     int symtabsize, unsigned int checksum) {
	union {
		struct panic_log_header plh;
		u8 sector[512];
	} secbuf;
	memset(secbuf.sector, 0, 512);
	secbuf.plh.syslogsize=syslogsize;
	secbuf.plh.plogsize=plogsize;
	secbuf.plh.symtabsize=symtabsize;
	secbuf.plh.checksum=checksum;
	memcpy(secbuf.plh.signature, PL_SIGNATURE, 11);
	strncpy(secbuf.plh.banner, linux_banner, PL_MAX_BANNER_SIZE);
	return pio_write_buffer_to_swap(secbuf.sector, 512, 0, NULL);
}

int do_panic_log(struct notifier_block *self, unsigned long command, void *ptr)
{
	char *syslog_data[4];
	static int panic_logged=0;
	unsigned long flags;
	save_and_cli(flags);

	if(panic_logged == 0) {
		int wrsects, plogsects, symsects;
		u32 csum=0;

		panic_logged=1; /* This only happens once */
		get_syslog_data(syslog_data);

		/* Write out syslog offset by one */
		wrsects=pio_write_buffer_to_swap(syslog_data[0],
				   syslog_data[1] - syslog_data[0], 1, &csum);
		if(wrsects < 0)
			return wrsects;

		/* Write out plog after syslog */
		plogsects = pio_write_plog(wrsects + 1, &csum);
		if(plogsects < 0)
			return plogsects;

		wrsects += plogsects;
		symsects = pio_write_syms(wrsects + 1, &csum);
		wrsects = pio_write_header(syslog_data[1] - syslog_data[0],
					   plogsects * 512, symsects * 512,
					   csum);
		if(wrsects < 0)
			return wrsects;
		
		printk("Panic logged\n");
	}
	restore_flags(flags);
	return NOTIFY_DONE;
}

static struct notifier_block panic_log_block = { do_panic_log, NULL, 10 };

static int __init panic_log_init(void) {
#ifdef CONFIG_TIVO_DEVEL
	/* Register /proc/uploadpanic so we get network feedback */
	struct proc_dir_entry *uploadpanic;
	/* Checked by rc.CheckPanicLog.sh */
	uploadpanic = create_proc_entry("uploadpanic", 0, 0);
#endif
	notifier_chain_register(&panic_notifier_list, &panic_log_block);
	printk("Kernel Panic Logger registered\n");
	return 0;
}

module_init(panic_log_init);
