/*
 * Giant Slalom data-structure definitions
 */

#include <linux/list.h>

enum gs_worker_state {
  GS_WORKER_SETUP = 0,
  GS_WORKER_PIO_TO_SDRAM,
  GS_WORKER_DMA_TO_SDRAM,
  GS_WORKER_DO_PSW_DMA_BURST,
  GS_WORKER_DETACH_COMMAND,
  GS_WORKER_QUEUE_FOR_ATAPI,
  GS_WORKER_QUEUE_FOR_GATE,
  GS_WORKER_INIT_ATAPI,
  GS_WORKER_SEND_PACKET,
  GS_WORKER_DMA_DATA,
  GS_WORKER_CHECK_DMA_RESULTS,
  GS_WORKER_PIO_DATA,
  GS_WORKER_CHECK_PIO_RESULTS,
  GS_WORKER_PIO_FROM_SDRAM,
  GS_WORKER_DMA_FROM_SDRAM,
  GS_WORKER_DO_PSR_DMA_BURST,
  GS_WORKER_LOOPBACK_DMA_DONE,
  GS_WORKER_POSTPROCESS,
  GS_WORKER_INIT_FOR_REQUEST_SENSE,
  GS_WORKER_SEND_REQUEST_SENSE,
  GS_WORKER_TRANSFER_SENSE_DATA,
  GS_WORKER_DONE,
  GS_WORKER_SETUP_RESET,
  GS_WORKER_RESET_DONE,
  GS_WORKER_SETUP_FL,
  GS_WORKER_FL_DATA_BURST,
  GS_WORKER_SETUP_DAC,
  GS_WORKER_DAC_DATA_BURST
};

struct gs_worker
{
  struct list_head link;
  unsigned long gsResourcesUsed;      /* mask of GS_RESOURCE_* */
  unsigned long gsResourcesNeeded;    /* mask of GS_RESOURCE_* */
  unsigned long gsInterruptsPending;  /* mask of GS_INT_* */
  unsigned long gsInterruptsFired;    /* mask of GS_INT_* */
  int sn;
  int doWriteback;
  int useAtapiDMA;
  int useHostDMA;
  int diagByte;
  enum gs_worker_state state;
  int sc_data_direction;
  int reportError;
  int reportHang;
  int go;
  int retries;
  int aborting;
  int timeout;
  int potentialDeadman;
  int deadman;
  int dwell;
  int awaitingPSR;
  int backgroundDMA;
  unsigned int transferSize;
  unsigned int gatedSize;
  unsigned int transferBurst;
  unsigned int transferRemaining;
  unsigned char sense_buffer[SCSI_SENSE_BUFFERSIZE];
  Scsi_Cmnd *scsiRequest;
  void *process;      /* fix this type */
  struct {
    unsigned char cmd[12];
    struct gs_sdram *buffer;
    unsigned char *transformBuffer;
    unsigned long transformBufferLength;
  } writeback;
  void (*endaction) (struct gs_worker *); /* worker-done callback */
  void *endactionToken;                  /* cookie for the callback */
};

#define GS_RESOURCE_ATAPI   0x00000001
#define GS_RESOURCE_HOSTDMA 0x00000002
#define GS_RESOURCE_SDRAM   0x00000004
#define GS_RESOURCE_FL      0x00000008
#define GS_RESOURCE_DAC     0x00000010
#define GS_RESOURCE_GATE    0x00000020

#define GS_MULTIPLE_GRABBERS   GS_RESOURCE_SDRAM

#define GS_INT_ATAPI        0x00000001
#define GS_INT_ATAPI_DMA    0x00000002
#define GS_INT_HOST_DMA     0x00000004
#define GS_INT_TIMER        0x00000008
#define GS_INT_PSR_VALID    0x00000010
#define GS_INT_PSR_STALL    0x00000020
#define GS_INT_PSW_STALL    0x00000040
#define GS_INT_BROADCOM_DMA 0x00000080
#define GS_INT_FL           0x00000100
#define GS_INT_DAC          0x00000200

struct gs_sdram 
{
  struct list_head link;
  struct gs_worker *worker;
  unsigned long gsOffset;
  unsigned long sizeRequested;
  unsigned long sizeAllocated; /* >= sizeRequested */
};

#define GS_SDRAM_MIN_ALLOC     2048
#define GS_SDRAM_ALLOC_MODULUS 4096

#define GS_SDRAM_GATE_THRESHOLD (6*1024*1024)
#define GS_SDRAM_GATE_BARRIER   (7*1024*1024)

#define GS_BRIDGE_BASE 0x1c100000
#define GS_WINDOW_BASE 0x1e000000

#define GS_BRIDGE ((volatile struct gs_bridge *) KSEG1ADDR(GS_BRIDGE_BASE))
#define GS_WRITE_WINDOW ((unsigned short *) KSEG1ADDR(GS_BRIDGE_BASE + 0x80000))
#define GS_READ_WINDOW ((unsigned short *) KSEG1ADDR(GS_BRIDGE_BASE + 0xC0000))

#define GS_BANK_SIZE   0x04000000

#define GS_OFFSET_CS0  0x00000000
#define GS_OFFSET_CS1  0x04000000
#define GS_OFFSET_CS2  0x08000000
#define GS_OFFSET_CS3  0x0C000000

#define GS_OFFSET_TRACON GS_OFFSET_CS0
#define GS_OFFSET_XCS1   GS_OFFSET_CS1
#define GS_OFFSET_CORE   GS_OFFSET_CS2
#define GS_OFFSET_SDRAM  GS_OFFSET_CS3

#define GS_OFFSET_IOS    GS_OFFSET_CS2+0x200000

#define GS_IRQ  BCM_IRQ_PCI_A2
#define DMA_IRQ BCM_IRQ_IODMA_DONE

#define GetGSBridgeAddress( addr ) ((volatile short *)KSEG1ADDR((addr)+GS_BRIDGE_BASE))
#define GetGSWindowAddress( addr ) ((volatile short *)KSEG1ADDR((addr)+GS_WINDOW_BASE))

struct gs_bridge 
{
  volatile unsigned short control;
  volatile unsigned short isr;
  volatile unsigned short imr;
  volatile unsigned short busTimeoutCounter;
  volatile unsigned short gsIndirectRegister;
  volatile unsigned short revision;
  volatile unsigned short revisionInverted;
};

#define GS_CONTROL_SOFT_RESET 0x0001
#define GS_CONTROL_PSW_RESET  0x0002
#define GS_CONTROL_PSR_RESET  0x0004
#define GS_CONTROL_GS_nRESET  0x0008
#define GS_CONTROL_PSR_ENABLE 0x0010
#define GS_CONTROL_BT_ENABLE  0x0020
#define GS_CONTROL_SPDIF      0x0040
#define GS_CONTROL_AUDCLOCK   0x0180
#define GS_CONTROL_DAC_MCLK   0x0200

#define GS_BRIDGE_INTR_PSR_DATA_VALID  0x0001
#define GS_BRIDGE_INTR_PSW_STALL       0x0002
#define GS_BRIDGE_INTR_PSR_STALL       0x0004
#define GS_BRIDGE_INTR_GS_STALL        0x0008
#define GS_BRIDGE_INTR_CS8420          0x0010
#define GS_BRIDGE_INTR_GS_XIRQ         0x0020
#define GS_BRIDGE_INTR_GS_XIRQS        0x0040

/*
  ATAPI register definitions
*/

#define GS_OFFSET_ATA_PRIMARY_COMMAND GS_OFFSET_CS2+0x1000000
#define GS_OFFSET_ATA_PRIMARY_CONTROL GS_OFFSET_CS2+0x1200000

#define GS_ATAPI_DATA_REGISTER      GS_OFFSET_ATA_PRIMARY_COMMAND+0
#define GS_ATAPI_ERROR_REGISTER     GS_OFFSET_ATA_PRIMARY_COMMAND+2
#define GS_ATAPI_FEATURE_REGISTER     GS_ATAPI_ERROR_REGISTER
#define GS_ATAPI_IREASON_REGISTER   GS_OFFSET_ATA_PRIMARY_COMMAND+4
/* no use of                        GS_OFFSET_ATA_PRIMARY_COMMAND+6 */
#define GS_ATAPI_BCOUNTL_REGISTER   GS_OFFSET_ATA_PRIMARY_COMMAND+8
#define GS_ATAPI_BCOUNTH_REGISTER   GS_OFFSET_ATA_PRIMARY_COMMAND+10
#define GS_ATAPI_SELECT_REGISTER    GS_OFFSET_ATA_PRIMARY_COMMAND+12
#define GS_ATAPI_COMMAND_REGISTER   GS_OFFSET_ATA_PRIMARY_COMMAND+14
#define GS_ATAPI_STATUS_REGISTER      GS_ATAPI_COMMAND_REGISTER

#define GS_ATAPI_CONTROL_REGISTER   GS_OFFSET_ATA_PRIMARY_CONTROL+12
#define GS_ATAPI_ALTSTATUS_REGISTER   GS_ATAPI_CONTROL_REGISTER

/*
 *	Bits of the interrupt reason register.
 */

#define GS_ATAPI_IREASON_COD	0x1 /* Information transferred is command */
#define GS_ATAPI_IREASON_IO	0x2 /* The device requests us to read */

/*
 * GS ATA Interrupt Mask
 */

#define GS_ATA_INTERRUPT_MASK_REGISTER GS_OFFSET_CORE+0x92

#define GS_ATA_INTERRUPT_MASK    0x0001
#define GS_ATA_INTERRUPT_CLEAR   0x0004

/*
  GS registers to control ATA primary DMA engine
*/

#define GS_ATA_DMA_APDC    GS_OFFSET_CORE+0x80
#define GS_ATA_DMA_IAPDC   GS_OFFSET_CORE+0x82
#define GS_ATA_DMA_APDB    GS_OFFSET_CORE+0x84
#define GS_ATA_DMA_APNB    GS_OFFSET_CORE+0x86
#define GS_ATA_DMA_APMA    GS_OFFSET_CORE+0x88
#define GS_ATA_DMA_APST    GS_OFFSET_CORE+0x8A
#define GS_ATA_DMA_APCT    GS_OFFSET_CORE+0x8C

/*
  GS registers to control host->SDRAM DMA
*/

#define GS_WRITE_DMA_WADC    GS_OFFSET_CORE+0x00
#define GS_WRITE_DMA_IWADC   GS_OFFSET_CORE+0x02
#define GS_WRITE_DMA_WADB    GS_OFFSET_CORE+0x04
#define GS_WRITE_DMA_WANB    GS_OFFSET_CORE+0x06
#define GS_WRITE_DMA_WAMA    GS_OFFSET_CORE+0x08
#define GS_WRITE_DMA_WAST    GS_OFFSET_CORE+0x0A
#define GS_WRITE_DMA_WACT    GS_OFFSET_CORE+0x0C

/*
  GS registers to control SDRAM->host DMA
*/

#define GS_READ_DMA_RADC    GS_OFFSET_CORE+0x60
#define GS_READ_DMA_IRADC   GS_OFFSET_CORE+0x62
#define GS_READ_DMA_RADB    GS_OFFSET_CORE+0x64
#define GS_READ_DMA_RANB    GS_OFFSET_CORE+0x66
#define GS_READ_DMA_RAMA    GS_OFFSET_CORE+0x68
#define GS_READ_DMA_RAST    GS_OFFSET_CORE+0x6A
#define GS_READ_DMA_RACT    GS_OFFSET_CORE+0x6C

/*
  GS registers to control DMA burst timings...
*/

#define GS_DMA_BANK0_CTL    GS_OFFSET_CORE+0x388
#define GS_DMA_BANK1_CTL    GS_OFFSET_CORE+0x38A
#define GS_DMA_BANK2_CTL    GS_OFFSET_CORE+0x38C
#define GS_DMA_BANK3_CTL    GS_OFFSET_CORE+0x383
#define GS_DMA_BANK4_CTL    GS_OFFSET_CORE+0x390
#define GS_DMA_BANK5_CTL    GS_OFFSET_CORE+0x392
#define GS_DMA_BANK6_CTL    GS_OFFSET_CORE+0x394
#define GS_DMA_BANK7_CTL    GS_OFFSET_CORE+0x396

/*
  Diagnostic and self-test registers in TRACON
*/

#define GS_TRACON_BANK0_BLOCK_IN_ENABLE       GS_OFFSET_TRACON+0x2F6
#define GS_TRACON_BANK0_BLOCK_OUT_ENABLE      GS_OFFSET_TRACON+0x2F8
#define GS_TRACON_BANK0_JTAG_ENABLE           GS_OFFSET_TRACON+0x308
#define GS_TRACON_BANK0_LOAD_ENABLE           GS_OFFSET_TRACON+0x320

#define GS_TRACON_BANK1_BLOCK_IN_ENABLE       GS_OFFSET_TRACON+0x2FA
#define GS_TRACON_BANK1_BLOCK_OUT_ENABLE      GS_OFFSET_TRACON+0x2FC
#define GS_TRACON_BANK1_JTAG_ENABLE           GS_OFFSET_TRACON+0x306
#define GS_TRACON_BANK1_LOAD_ENABLE           GS_OFFSET_TRACON+0x324

#define GS_TRACON_INTERFACE_SENSE             GS_OFFSET_TRACON+0x2DC

#define GS_TRACON_MISC_REG_380                GS_OFFSET_TRACON+0x380
#define GS_TRACON_MISC_REG_382                GS_OFFSET_TRACON+0x382
#define GS_TRACON_MISC_REG_384                GS_OFFSET_TRACON+0x384
#define GS_TRACON_MISC_REG_386                GS_OFFSET_TRACON+0x386
#define GS_TRACON_MISC_REG_388                GS_OFFSET_TRACON+0x388
#define GS_TRACON_MISC_REG_38A                GS_OFFSET_TRACON+0x38A
#define GS_TRACON_MISC_REG_38C                GS_OFFSET_TRACON+0x38C
#define GS_TRACON_MISC_REG_38E                GS_OFFSET_TRACON+0x38C

/*
  Serial I/O registers for SPI to FL display
*/

#define GS_FL_CDATA             GS_OFFSET_IOS+0x450
#define GS_FL_CDATA_END         GS_OFFSET_IOS+0x46F
#define GS_FL_CCTL1             GS_OFFSET_IOS+0x470
#define GS_FL_CCTL2             GS_OFFSET_IOS+0x472
#define GS_FL_CLATCH            GS_OFFSET_IOS+0x474
#define GS_FL_CBYTNM            GS_OFFSET_IOS+0x476
#define GS_FL_CCKSEL            GS_OFFSET_IOS+0x478
#define GS_FL_CINTSL            GS_OFFSET_IOS+0x47A
#define GS_FL_CINT              GS_OFFSET_IOS+0x47C

/*
  Serial I/O registers for SPI to DAC
*/

#define GS_DAC_DDATA             GS_OFFSET_IOS+0x480
#define GS_DAC_DDATA_END         GS_OFFSET_IOS+0x49F
#define GS_DAC_DCTL1             GS_OFFSET_IOS+0x4A0
#define GS_DAC_DCTL2             GS_OFFSET_IOS+0x4A2
#define GS_DAC_DLATCH            GS_OFFSET_IOS+0x4A4
#define GS_DAC_DBYTNM            GS_OFFSET_IOS+0x4A6
#define GS_DAC_DCKSEL            GS_OFFSET_IOS+0x4A8
#define GS_DAC_DINTSL            GS_OFFSET_IOS+0x4AA
#define GS_DAC_DINT              GS_OFFSET_IOS+0x4AC

/*
  Serial I/O bank control registers
*/

#define GS_SERIAL_CTRL           GS_OFFSET_IOS+0x4B0
#define GS_SERIAL_MODE           GS_OFFSET_IOS+0x4B2
#define GS_SERIAL_WDATA          GS_OFFSET_IOS+0x4B4
#define GS_SERIAL_RDATA          GS_OFFSET_IOS+0x4B6
#define GS_SERIAL_ACK            GS_OFFSET_IOS+0x4B8
#define GS_SERIAL_RATE           GS_OFFSET_IOS+0x4BA
#define GS_SERIAL_INTRPT         GS_OFFSET_IOS+0x4BC
#define GS_SERIAL_STATE          GS_OFFSET_IOS+0x4BE
#define GS_SERIAL_IODIR1         GS_OFFSET_IOS+0x4C0
#define GS_SERIAL_IODATA1        GS_OFFSET_IOS+0x4C2
#define GS_SERIAL_IODIR2         GS_OFFSET_IOS+0x4C4
#define GS_SERIAL_IODATA2        GS_OFFSET_IOS+0x4C6
#define GS_SERIAL_SELSER         GS_OFFSET_IOS+0x4C8
#define GS_SERIAL_IODIR3         GS_OFFSET_IOS+0x4CA
#define GS_SERIAL_IODATA3        GS_OFFSET_IOS+0x4CC
#define GS_SERIAL_SELI2C         GS_OFFSET_IOS+0x4CE

/*
  GS Interrupt Factor registers
*/

#define GS_INTERRUPT_FACTOR_A   GS_OFFSET_CORE+0x126
#define GS_INTERRUPT_FACTOR_B   GS_OFFSET_CORE+0x128
#define GS_INTERRUPT_S_FACTOR_A GS_OFFSET_CORE+0x12A
#define GS_INTERRUPT_S_FACTOR_B GS_OFFSET_CORE+0x12C

#define GS_INTERRUPT_S_READ_DMA_FIFO       0x00200000
#define GS_INTERRUPT_S_READ_DMA            0x00100000
#define GS_INTERRUPT_S_WRITE_DMA_PACK      0x00020000
#define GS_INTERRUPT_S_WRITE_DMA           0x00010000
#define GS_INTERRUPT_S_ATA_SECONDARY       0x00008000
#define GS_INTERRUPT_S_ATA_SECONDARY_PACK  0x00004000
#define GS_INTERRUPT_S_ATA_SECONDARY_DMA   0x00002000
#define GS_INTERRUPT_S_ATA_PRIMARY         0x00000800
#define GS_INTERRUPT_S_ATA_PRIMARY_PACK    0x00000400
#define GS_INTERRUPT_S_ATA_PRIMARY_DMA     0x00000200
#define GS_INTERRUPT_S_SERIAL_D            0x00000100
#define GS_INTERRUPT_S_SERIAL_C            0x00000080
#define GS_INTERRUPT_S_SERIAL_B            0x00000040
#define GS_INTERRUPT_S_SERIAL_A            0x00000020
#define GS_INTERRUPT_S_I2C                 0x00000010
#define GS_INTERRUPT_S_TIMESTAMP           0x00000008
#define GS_INTERRUPT_S_TRACON              0x00000004
#define GS_INTERRUPT_S_DIAG                0x00000002
#define GS_INTERRUPT_S_FIRMWARE_LOAD       0x00000001

